//OpenGL Alpha Masking, Blending & Fog example

#include <GL/glut.h>
#include "tga.h"

/* texture defines */
#define TOTAL_NUMBER_OF_TEXTURES 3
#define BARK 0
#define LEAVES 1
#define FLOOR 2

/* initial lists */
#define LIST_SCENE 1
#define LIST_CLEANUP 2
#define LIST_INITIALISE 3
#define LIST_SETUP 4

/* internal storage for texture data */
GLuint theTexture[TOTAL_NUMBER_OF_TEXTURES];

/* texture filename list */
char* textureFilenames[TOTAL_NUMBER_OF_TEXTURES] = {"bark.tga",
													"leaves.tga",
													"floor.tga"};

/*
=================
imageLoader
-loads our TGA data
=================
*/
bool imageLoader(char *TexName, int currentTexture) 
{
	/* image loading type */
	TGAImg Img;

	/* loeads the texture */
	if(Img.Load(TexName)!=IMG_OK) 
		return false;

	/* the texture handle is set to the current loaded texture */
	glBindTexture(GL_TEXTURE_2D, theTexture[currentTexture]);

	/* setup the texture if it is a standard solid 24bit image (no transparency) */
	if(Img.GetBPP() == 24) 
	{
		/* specifies the solid two-dimensional texture */
		glTexImage2D(GL_TEXTURE_2D, 0, 3, Img.GetWidth(), Img.GetHeight(), 0, GL_RGB,GL_UNSIGNED_BYTE, Img.GetImg());
	} 
	/* else setup the texture if it is a 32bit image with an alpha channel */
	else if(Img.GetBPP()==32) 
	{
		/* specifies the transparent two-dimensional texture */
		glTexImage2D(GL_TEXTURE_2D, 0, 4, Img.GetWidth(),Img.GetHeight(), 0, GL_RGBA, GL_UNSIGNED_BYTE, Img.GetImg());
	} 
	else 
	{
		return false;
	}

	/* sets the texture parameters - for reference, consult Prog 8.6's discussion */
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

	return true;
}

/*
=================
textureLoader
-calls imageLoader for every TGA file
=================
*/
void textureLoader() 
{
	/* deals with the generation of texture names  */
	glGenTextures(TOTAL_NUMBER_OF_TEXTURES, theTexture);

	/* calls imageLoader for the current texture */
	for (int k=0; k < TOTAL_NUMBER_OF_TEXTURES; k++) 
	{
		imageLoader(textureFilenames[k], k);
	}
}

/*
=================
makePosOfViewerList
-setup the viewer's position in the world
=================
*/
void makePosOfViewerList()
{
	/* cleate the display list LIST_SETUP 
	   - everything up to glEndList is placed in this list 
	   - GL_COMPILE: only compile the commands into the list 
	   - GL_COMPILE_AND_EXECUTE: execute the commands as compiled into list */
	glNewList(LIST_SETUP, GL_COMPILE);
		/* clear frame buffer and z-buffer - thus clearing the screen */
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		/* pushes current matrix stack down by one */
		glPushMatrix();
			/*  translate (move) camera -4 units (down) on y-axis */
			glTranslatef(0.0, -4, 0);
			/* specifies the Modelview matrix as the current matrix */
			glMatrixMode(GL_MODELVIEW);
			/* move to the center of the screen */
			glLoadIdentity();
		/* pops current matrix stack */
		glPopMatrix();
	/* indicates the end of the list */
	glEndList();
}

/*
=================
display
-setup the fog parameters and our LIST_INITIALISE display list
=================
*/
void display()
{
	/* set the clear colour - red, green, blue, alpha*/
	glClearColor(1.0, 1.0, 1.0, 0.0);

	/* specify the fog colour as off-white */
	float fog_colour[] = {0.8, 0.8, 0.8, 1.0};
	/* set the fog mode to exponential */
	glFogf(GL_FOG_MODE, GL_EXP);
	/* specifies the fog density */
	glFogf(GL_FOG_DENSITY, 0.03);
	/* sets the fog colour */
	glFogfv(GL_FOG_COLOR, fog_colour);
	/* call the LIST_INITIALISE display list */
	glCallList(LIST_INITIALISE);
	/* swap the buffers for double buffering */
	glutSwapBuffers();
}

/*
=================
drawScene
-function responsible for the drawing of the trees, leaves & floor
=================
*/
void drawScene()
{
	/* moves -5 units (left) on the x-axis and -25 units (down) on y-axis */
	glTranslatef(-5, 25, 0);

	/* the first tree showing blending */
	glPushMatrix();
		/* create the quadrics object */
		GLUquadricObj* obj;
		obj = gluNewQuadric();
		
		/* enables the creation of a texture bound to a texture target */
		glBindTexture(GL_TEXTURE_2D, theTexture[BARK]);
		
		/* sets the quadrics' draw style */
		gluQuadricDrawStyle(obj, GLU_FILL);
		/* affects the overall shading, either flat or smooth */
		gluQuadricNormals(obj, GLU_SMOOTH);
		/* sets texturing of the quadric enabled or disabled */
		gluQuadricTexture(obj, GLU_TRUE);

		/* gluCylinder(quadric object, cylinder's base radius, cylinder's top radius, 
	                   height of cylinder, longitude subdivisions, latitude subdivisions */
        gluCylinder(obj, 2.0, 1.5, 30.0, 12, 5);
		
		glPushMatrix();
			/* enable OpenGL blending */
			glEnable(GL_BLEND);

			/* define the blending operation */
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

			/* binds the leaf texture */
			glBindTexture(GL_TEXTURE_2D, theTexture[LEAVES]);
			
			/* enable hidden-surface removal */
			glEnable(GL_DEPTH_TEST);
			
			/* base cone of the tree - for the leaves */
			glTranslatef(0,0,5);		
			gluCylinder(obj, 7.0, 0.0, 10, 12, 5);
			
			/* middle cone of the tree - for the leaves */
			glPushMatrix();
				glTranslatef(0,0,5);
				gluCylinder(obj, 5.0, 0.0, 7, 12, 5);
			glPopMatrix();
			
			/* top cone of the tree - for the leaves */
			glPushMatrix();
				glTranslatef(0,0,5);
				gluCylinder(obj, 3.0, 0.0, 30, 12, 5);
			glPopMatrix();
		glPopMatrix();
	glPopMatrix();

	/* the second tree showing alpha masking */
	glPushMatrix();
		/* disable blending for alpha masking */
		glDisable(GL_BLEND);

		glTranslatef(10, 3, 0);
		
		glBindTexture(GL_TEXTURE_2D, theTexture[BARK]);
		gluCylinder(obj, 2.0, 1.5, 30.0, 12, 5);
			
		glPushMatrix();
			/* enable alpha masking */
			glEnable(GL_ALPHA_TEST);

			/* set the alpha test function */
			glAlphaFunc(GL_GREATER, 0.45);

			glBindTexture(GL_TEXTURE_2D, theTexture[LEAVES]);
			glEnable(GL_DEPTH_TEST);
			
			glTranslatef(0,0,5);
			gluCylinder(obj, 7.0, 0.0, 10, 12, 5);
			
			glPushMatrix();
				glTranslatef(0,0,5);
				gluCylinder(obj, 5.0, 0.0, 7, 12, 5);
			glPopMatrix();
			
			glPushMatrix();
				glTranslatef(0,0,5);
				gluCylinder(obj, 3.0, 0.0, 30, 12, 5);	
			glPopMatrix();
		glPopMatrix();
	glPopMatrix();

	/* the third tree for illustration of fog falloff */
	glPushMatrix();
		glTranslatef(6, 35, 0);

		glBindTexture(GL_TEXTURE_2D, theTexture[BARK]);
			gluCylinder(obj, 2.0, 1.5, 30.0, 12, 5);
		
			glPushMatrix();
				/* enable alpha masking */
				glEnable(GL_ALPHA_TEST);
				/* set the alpha test function */
				glAlphaFunc(GL_GREATER, 0.45);

				glBindTexture(GL_TEXTURE_2D, theTexture[LEAVES]);
					
				glTranslatef(0,0,5);
				gluCylinder(obj, 7.0, 0.0, 5, 12, 5);
				
				glPushMatrix();
					glTranslatef(0,0,5);
					gluCylinder(obj, 5.0, 0.0, 7, 12, 5);
				glPopMatrix();
				
				glPushMatrix();
					glTranslatef(0,0,5);
					gluCylinder(obj, 3.0, 0.0, 50, 12, 5);	
				glPopMatrix();
			glPopMatrix();
	glPopMatrix();

	/* draw the floor */
	glPushMatrix();
		glTranslatef(6, 20, 0);
		glRotatef(90, 0, 1, 0);
	
		glBindTexture(GL_TEXTURE_2D, theTexture[FLOOR]);
		glBegin(GL_QUADS);
			glNormal3f(0,1,0);
			glTexCoord2f(0.0,0.0);
			glVertex3f(0.0, -50.0, -50.0);
			glTexCoord2f(0.0,1.0);
			glVertex3f(0.0, -50.0, 50.0);
			glTexCoord2f(1.0,1.0); //the bigger the higher the resulution
			glVertex3f(0.0, 50.0, 50.0);
			glTexCoord2f(1.0,0.0);
			glVertex3f(0.0, 50.0, -50.0);
		glEnd();
	glPopMatrix();
}

/*
=================
createSceneList
-defines a list, LIST_SCENE, storing the 3D scene
=================
*/
void makeSceneList()
{
	/* enable texturing mode */
	glEnable(GL_TEXTURE_2D);
	
	/* create the list, LIST_SCENE */
	glNewList(LIST_SCENE, GL_COMPILE);
		drawScene();
	glEndList();
}

/*
=================
setupLights
-specify the lights for our scene
=================
*/
void setupLights()
{	
	/* enables lighting  without it entire scene appears dark */
	glEnable(GL_LIGHTING);

	glShadeModel(GL_SMOOTH); //glShadeModel(GL_FLAT); //for flat shading

	/* activate light x (GL_LIGHTx) in the evaluation of the lighting equation 
	   - can be any # but logical numbering helps...*/
	glEnable(GL_LIGHT1);
	/* specify the colour of the light (white) */
	GLfloat diffuse_light_colour[] = {1.0f, 1.0f, 1.0f};
	/* specify the position of the light */
	GLfloat diffuse_light_position[] = {20.0, 20.0, -20.0, 1.0f};
	
	/* sets the light source parameters */
	glLightfv( GL_LIGHT1, GL_DIFFUSE, diffuse_light_colour);
	/* sets the light source position */
	glLightfv( GL_LIGHT1, GL_POSITION, diffuse_light_position);

	/* activate GL_LIGHT2 */
	glEnable(GL_LIGHT2);
	/* specify the colour of the light (white) */
	GLfloat diffuse_light_colour2[] = {1.0f, 1.0f, 1.0f};
	/* specify the position of the light */
	GLfloat diffuse_light_position2[] = {-20.0, 50.0, 20.0, 1.0f};
	
	/* sets the light source parameters */
	glLightfv( GL_LIGHT2, GL_DIFFUSE, diffuse_light_colour2);
	/* sets the light source position */
	glLightfv( GL_LIGHT2, GL_POSITION, diffuse_light_position2);

	/* activate GL_LIGHT3 */
	glEnable(GL_LIGHT3);
	/* specify the colour of the light (white) */
	GLfloat ambient_light_colour[] = {1.0f, 1.0f, 1.0f};
	/* specify the position of the light */
	GLfloat ambient_light_position[] = {0.0, 0.0, 0.0, 1.0f};
	
	/* sets the light source parameters */
	glLightfv( GL_LIGHT3, GL_AMBIENT, ambient_light_colour);
	/* sets the light source position */
	glLightfv( GL_LIGHT3, GL_POSITION, ambient_light_position);
}

/*
=================
setFogOn
-function used for enabling fog (key 'f')
=================
*/
void setFogOn()
{
	glEnable(GL_FOG);
}

/*
=================
setFogOff
-function used for disabling fog (key 'g')
=================
*/
void setFogOff()
{
	glDisable(GL_FOG);
}

/*
=================
keys
-Handles all normal keys input
=================
*/
void keys(unsigned char key, int x, int y)
{
	/* exit the program */
	if(key == 'q' || key == 'Q')
	{
		exit(-1);
	}

	/* enable fog */
	if(key == 'f' || key == 'F')
	{
		setFogOn();
	}

	/* disable fog */
	if(key == 'g' || key == 'G')
	{
		setFogOff();
	}
}

/*
=================
idle
-redisplay our window
=================
*/
void idle()
{	
	/* sets the current window for redisplay */
	glutPostRedisplay();
}

/*
=================
myInit
-handles basic initialisation and setup of the OpenGL parameters
=================
*/
int myInit()
{
	/* enable fog */
	setFogOn();

	/* setup the texture loader */
	textureLoader();

	/* setup the lights */
	setupLights();

	/* create our viewing position list */
	makePosOfViewerList();

	/* create our 3D scene's display list */
	makeSceneList();

	/* primary list used to setup our depth testing & 3D scene 
	   (called during display() function */
	glNewList(LIST_INITIALISE, GL_COMPILE);
		/* call list setting our viewing position (defined in makePosOfViewerList() */
		glCallList(LIST_SETUP);
		glPushMatrix();
			glRotatef(-90.0, 1.0, 0.0, 0.0);
			/* call 3D scene's list (defined in makeSceneList() */
			glCallList(LIST_SCENE);
		glPopMatrix();
		glMatrixMode(GL_PROJECTION);
		glPopMatrix();
	glEndList();
	
	return 1;
}

/*
================
reshape
-Rendered window's reshape function - manages changes to size of the window, correctly adjusting aspect ratio.
=================
*/
void reshape(int w, int h)
{
	/* sets the viewport with the first two parameters the lower-left corner of the viewport 
	   rectangle, in pixels and the last two, the width and height of the viewport */
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);

	/* specifies the Projection matrix as the current matrix */
	glMatrixMode(GL_PROJECTION);
	/* move to the center of the screen */
	glLoadIdentity();

	/* sets up the  perspective projection matrix:
	   - the first parameter specifies the field of view (fov), an angle in degrees 
	     (in the y-direction)
	   - the second parameter is aspect ratio determining fov 
		 (in the x-direction)
	   - the third parameter is the distance from the viewer to the near clipping plane
	   - the last parameter is the distance from the viewer to the far clipping plane */
	gluPerspective(45, (GLdouble)w / (GLdouble)h, 0.05, 100);
	/* specifies the Modelview matrix as the current matrix */
	glMatrixMode(GL_MODELVIEW);
}

/*
=================
main
=================
*/
int main(int argc, char **argv)
{
	/* initialize the GLUT library. */
	glutInit(&argc, argv);
	/* sets the initial display mode: double buffering and a depth buffer. */
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_ALPHA);
	/* sets default window size */
	glutInitWindowSize(640, 480);
	/* creates a top-level window with a caption */
	glutCreateWindow("...Fog, Alpha Masking & Blending...");
	/* reshape callback for the current window */
	glutReshapeFunc(reshape);
	/* displays our contents in the window */
	glutDisplayFunc(display);
	/* keys handling */
	glutKeyboardFunc(keys);
	/* sets the global idle callback for background processing/animation */
	glutIdleFunc(idle);
	myInit();
	/* enters the GLUT event processing loop - will handle any callback */
	glutMainLoop();
}
//EOF